/* Author: Domagoj Babic 2004.
 *
 * $Id: hcube.h,v 1.10 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#include "utils.h"
#include "datas.h"

// nxt & neg nodes:
// ptr  - pointer to next node
//   1  - termination node (needs to be explored)
//   0  - only valid for neg nodes for non-branching nodes
typedef struct hc_node_t {
    struct hc_node_t *nxt; // pos branch for branching nodes
    struct hc_node_t *neg;
    uint var;
} hc_node_t;

// termination node
#define TERM_ND (hc_node_t *)1
// termination macro
#define TERMINATE(x) (hc_node_t *)_umax((uint)x, (uint)1);

typedef struct hc_anchor_t {
    hc_node_t *nd;
    uint has_hcube;
    uint num_nodes;
    uint num_bnodes;
} hc_anchor_t;

typedef struct hc_support_t {
    uint varnum;
    int addedNodes;
    int addedBnodes;
    uint follow_all_branches;
    uint too_branchy;
    uint max_bnodes;
    uint conflict;
    uint branch_depth;
    // maximum ration of bnodes/nodes. if adding a new
    // branch would increase the ration of the hcube 
    // above the treshold, the algorithm will just free
    // everything below newly created branch.
    float treshold;
    stack_t *elim;
    stack_t *pbranch;
    hc_node_t *thc;
    table_t *ctable; // certificate table
    uint sat_manager_ref;
    uint *indices;
    int (*prifunc)(uint, uint);
    unsigned char *stamps;
} hc_support_t;

hc_support_t *initHcSupport(uint size, uint sm_ref, int (*prifunc)(uint, uint));
void freeHcSupport(hc_support_t *hcs);
hc_node_t *getHcNode(hc_support_t *hcs, uint var);
inline uint hc_is_bnode(hc_node_t *x);
void freeHcNode(hc_support_t *hcs, hc_node_t *n);
inline uint hc_too_branchy(hc_support_t *hcs, hc_anchor_t *ha);
void freeOneHcNode(hc_support_t *hcs, hc_node_t *n);
void resetHcube(hc_support_t *hcs, hc_anchor_t *ha);
void resetHcStacks(hc_support_t *hcs);
inline void hc_link(hc_node_t *n1, hc_node_t *n2);
inline void append(hc_node_t *n1, hc_node_t *n2);
hc_node_t *follow_branch(hc_support_t *hcs, hc_node_t *n, flexarray_t *c);
hc_node_t *follow_path(hc_support_t *hcs, hc_node_t *n, flexarray_t *c);
void common_path(hc_support_t *hcs, hc_anchor_t *hc, flexarray_t *c);
uint path_intersect(hc_support_t *hcs, hc_anchor_t **hc);

hc_support_t *initHcSupport(uint size, uint sm_ref, int (*prifunc)(uint, uint)) {
    hc_support_t *hcs = xcalloc(1, sizeof(hc_support_t));
    hcs->varnum = size;
    hcs->elim = initStack(16);
    hcs->pbranch = initStack(16);
    hcs->ctable = initTable(size + 1, 1);
    hcs->indices = xmalloc((size + 1) * sizeof(uint));
    hcs->stamps = (unsigned char *)xmalloc(size);
    hcs->sat_manager_ref = sm_ref;
    hcs->prifunc = prifunc;
    return hcs;
}

void resetHcSupport(hc_support_t *hcs) {
    hcs->addedNodes = 0;
    hcs->addedBnodes = 0;
    assert(stackEmpty(hcs->elim));
    assert(stackEmpty(hcs->pbranch));
}

void freeHcSupport(hc_support_t *hcs) {
    stackFree(hcs->elim);
    stackFree(hcs->pbranch);
    freeTable(hcs->ctable);
    free(hcs->indices);
    free(hcs->stamps);
    free(hcs);
}

inline unsigned char getStamp(hc_support_t *hcs, uint idx) {
    assert(idx < hcs->varnum);
    return hcs->stamps[idx] & TRUE;
}

inline void markStamp(hc_support_t *hcs, uint idx) {
    unsigned char *stamps = hcs->stamps;
    stamps[idx] |= TRUE;
}

inline void saveStamps(hc_support_t *hcs, uint size) {
    uint i;
    unsigned char *stamps = hcs->stamps;
    for (i = 0; i < size; i++) {
        assert((unsigned char)(stamps[i]) < (unsigned char)128);
        stamps[i] = LSB(stamps[i]) | (stamps[i] << 1);
    }
}

inline void revertStamps(hc_support_t *hcs, uint size) {
    uint i;
    unsigned char *stamps = hcs->stamps;
    for (i = 0; i < size; i++) {
        stamps[i] >>= 1;
    }
}    

inline void clearStamps(hc_support_t *hcs, uint size) {
    assert(size < hcs->varnum);
    memset(hcs->stamps, 0, size);
}

hc_node_t *getHcNode(hc_support_t *hcs, uint var) {
    hc_node_t *n = xcalloc(1, sizeof(hc_node_t));
    assert(var < ((hcs->varnum + 1) * 2) && var > 0);
    n->var = var;
    n->nxt = TERM_ND;
    hcs->addedNodes++;
    return n;
}

inline uint hc_is_bnode(hc_node_t *x) {
    assert(x != NULL);
    assert((uint)x > 1);
    return (uint)x->neg;
}

void freeHcNode(hc_support_t *hcs, hc_node_t *n) {
    if ((uint)n < 2) {
        return;
    } else if (hc_is_bnode(n)) {
        freeHcNode(hcs, TRIM(n->nxt));
        freeHcNode(hcs, TRIM(n->neg));
        hcs->addedBnodes--;
    } else {
        freeHcNode(hcs, n->nxt);
        hcs->addedNodes--;
    }
    free(n);
}

void freeHcAnchor(hc_support_t *hcs, hc_anchor_t *ha) {
    freeHcNode(hcs, ha->nd);
    free(ha);
}

inline void hc_update_cnts(hc_support_t *hcs, hc_anchor_t *ha) {
    ha->num_nodes += hcs->addedNodes;
    ha->num_bnodes += hcs->addedBnodes;
    resetHcSupport(hcs);
}

inline uint hc_too_branchy(hc_support_t *hcs, hc_anchor_t *ha) {
    uint sum = ha->num_nodes + ha->num_bnodes;
    return (float)ha->num_bnodes/(sum) >= hcs->treshold || 
        ha->num_bnodes >= hcs->max_bnodes;
}

void freeOneHcNode(hc_support_t *hcs, hc_node_t *n) {
    assert(n != NULL);
    if (hc_is_bnode(n)) {
        hcs->addedBnodes--;
    } else {
        hcs->addedNodes--;
    }
    free(n);
}

void resetHcStacks(hc_support_t *hcs) {
    // discard all potentially branching nodes
    while (!stackEmpty(hcs->pbranch)) {
        hc_node_t *nd = (hc_node_t *)stackPop(hcs->pbranch);
        freeOneHcNode(hcs, nd);
    }
    // discard all elimated nodes
    while (!stackEmpty(hcs->elim)) {
        hc_node_t *nd = (hc_node_t *)stackPop(hcs->elim);
        freeOneHcNode(hcs, nd);
    }
}

inline void hc_link(hc_node_t *n1, hc_node_t *n2) {
    assert(n1 || n2);
    if (!n2) {
        n1->nxt = TERM_ND;
    } else {
        n1->nxt = n2;
    }
}

inline void append(hc_node_t *n1, hc_node_t *n2) {
    assert(n1 != NULL);
    assert(n2 != NULL);
    n1->nxt = n2;
    n2->nxt = TERM_ND;
}

// copies the list of nodes starting with nd, and inverts
// all literals
void invListCopy(hc_support_t *hcs, hc_node_t *nd, uint num, 
        hc_node_t **first, hc_node_t **last) {
    hc_node_t *f = NULL, *l = NULL;
    while (nd && num--) {
        hc_node_t *x = getHcNode(hcs, nd->var ^ 1);
        if (l) {
            l->nxt = x;
            l = x;
        } else {
            f = l = x;
        }
        nd = nd->nxt;
    }
    if (l) {
        //l->nxt = TERM_ND;
        assert(!l->neg);
        assert((uint)l->nxt >= 1);
        *first = f;
        *last = l;
    }
}

// returns the number of enlisted pbranch nodes
uint enlistStacks(hc_support_t *hcs, hc_node_t **first, hc_node_t **last) {
    hc_node_t *f = NULL, *l = NULL;
    stack_t *s;
    uint pbranch_nodes = 0;
    s = hcs->pbranch;
    while (!stackEmpty(s)) {
        hc_node_t *x = (hc_node_t *)stackPop(s);
        if (!x) {
            continue;
        }
        pbranch_nodes++;
        if (l) {
            l->nxt = x;
            l = x;
        } else {
            f = l = x;
        }
    }
    s = hcs->elim;
    while (!stackEmpty(s)) {
        hc_node_t *x = (hc_node_t *)stackPop(s);
        if (l) {
            l->nxt = x;
            l = x;
        } else {
            f = l = x;
        }
    }
    if (l) {
        l->nxt = TERM_ND;
        assert(!l->neg);
        *first = f;
        *last = l;
    }
    return pbranch_nodes;
}

void stacksDeepFill(hc_support_t *hcs, uint *elim, uint elim_sz, 
        uint *pbranch, uint pbranch_sz) {
    uint i;
    for (i = 0; i < elim_sz; i++) {
        hc_node_t *n = (hc_node_t *)elim[i];
#ifdef DEBUG
        uint v = getVar(n->var);
        assert(v <= hcs->varnum && v > 0);
#endif
        stackPush(hcs->elim, (Item)getHcNode(hcs, n->var));
    }
    for (i = 0; i < pbranch_sz; i++) {
        hc_node_t *n = (hc_node_t *)pbranch[i];
#ifdef DEBUG
        uint v = getVar(n->var);
        assert(v <= hcs->varnum && v > 0);
#endif
        stackPush(hcs->pbranch, (Item)getHcNode(hcs, n->var));
    }
}

inline int find_in_certificate(hc_support_t *hcs, uint x) {
    uint v = getVar(x);
    assert(v <= hcs->varnum && v > 0);
    return (checkTable(hcs->ctable, v) < 0) ? -1 : (int)hcs->indices[v];
}

hc_node_t *follow_branch(hc_support_t *hcs, hc_node_t *n, flexarray_t *c) {
	int i;
    assert(hc_is_bnode(n));
    i = find_in_certificate(hcs, n->var);
    if (i < 0) { // branch var not present in the certificate
        hcs->branch_depth++;
        // if in agressive pruning mode or if there's no more places
        // to shift stored stamp bits (sizeof(char) = 8 bits) eliminate
        // the whole branch
        if (!hcs->follow_all_branches || hcs->branch_depth >= 7) {
            freeHcNode(hcs, n);
            n = NULL;
        } else { // follow both branches
            uint csz = getFlexArraySize(c);
            uint elim_sz, pbranch_sz, nelim_sz, npbranch_sz;
            uint *elim = NULL, *pbranch = NULL;
#ifdef DEBUG
            uint tmp_elim_sz, tmp_pbranch_sz;
#endif

            // handle positive branch
            saveStamps(hcs, csz);
            stackDump(hcs->elim, &elim, &elim_sz);
            stackDump(hcs->pbranch, &pbranch, &pbranch_sz);
            assert((uint)n->nxt >= 1);
            n->nxt = follow_path(hcs, n->nxt, c);
            assert((uint)n->nxt >= 1);
            revertStamps(hcs, csz);

            nelim_sz = stackSize(hcs->elim);
            npbranch_sz = stackSize(hcs->pbranch);

            assert(nelim_sz == 0 || nelim_sz == elim_sz);
            assert(npbranch_sz == 0 || npbranch_sz == pbranch_sz);

            if ((nelim_sz + npbranch_sz) == 0) {
                stacksDeepFill(hcs, elim, elim_sz, pbranch, pbranch_sz);
            }

            if (elim) free(elim);
            if (pbranch) free(pbranch);
            
            // handle negative branch
            assert((uint)n->neg >= 1);
            n->neg = follow_path(hcs, n->neg, c);
            assert((uint)n->neg >= 1);
            
#ifdef DEBUG
            nelim_sz = stackSize(hcs->elim);
            npbranch_sz = stackSize(hcs->pbranch);

            assert(nelim_sz == 0 || nelim_sz == elim_sz);
            assert(npbranch_sz == 0 || npbranch_sz == pbranch_sz);
#endif

            if ((uint)n->nxt == 1 && (uint)n->neg == 1) {
                freeOneHcNode(hcs, n);
                n = NULL;
            }

#ifdef DEBUG
            tmp_elim_sz = stackSize(hcs->elim);
            tmp_pbranch_sz = stackSize(hcs->pbranch);
            assert(tmp_elim_sz == 0 || tmp_elim_sz == elim_sz);
            assert(tmp_pbranch_sz == 0 || tmp_pbranch_sz == pbranch_sz);
#endif
        }
        hcs->branch_depth--;
    } else { // branching variable found in certificate
        uint *ca = getFlexArray(c);
        uint lit = ca[i];
        hc_node_t *first = NULL, *last = NULL, *tmp = NULL;
        // trail is the path that will be followed
        hc_node_t **trail = (LSB(lit)) ? &(n->nxt) : &(n->neg);
        assert(i >= 0);
        // mark that the literal has been seen on the path 
        markStamp(hcs, (uint)i);

        // add eliminated and potentially branching nodes to the list
        (void)enlistStacks(hcs, &first, &last);
        // TODO: try enlisting only eliminated nodes, or 
        // both eliminated and branch nodes, but without emptying
        // branch node stack
        
        if (first) {
            // other is the branch that is not followed
            hc_node_t **other = (LSB(lit)) ? &(n->neg) : &(n->nxt);
            assert((uint)last >= 1);
            assert((uint)*other >= 1);
            last->nxt = *other;
            *other = first;
        }
        assert(stackEmpty(hcs->elim));
        assert(stackEmpty(hcs->pbranch));
        assert((uint)*trail >= 1);
        tmp = follow_path(hcs, *trail, c);
        assert((uint)tmp >= 1);
        *trail = tmp;
        
        assert(n->nxt && n->neg);
        if ((uint)n->nxt == 1 && (uint)n->neg == 1) {
            freeOneHcNode(hcs, n);
            n = NULL;
        }
    }
    return n;
}

hc_node_t *pickBestBnode(hc_support_t *hcs) {
    int i, maxi = 0, maxp = 0;
    hc_node_t *nd;
    stack_t *s = hcs->pbranch;
    assert(!stackEmpty(s));
    for (i = (int)stackSize(hcs->pbranch) - 1; i >= 0; i--) {
        hc_node_t *n = (hc_node_t *)stackReadAt(s, i);
        uint v;
        assert(n != NULL);
        assert(n != TERM_ND);
        v = getVar(n->var);
        if (hcs->prifunc(hcs->sat_manager_ref, v) > maxp) {
            maxi = i;
        }
    }
    nd = (hc_node_t *)stackReadAt(s, maxi);
    stackWriteAt(s, maxi, (Item)0);
    assert(nd != NULL);
    assert(nd != TERM_ND);
    assert(getVar(nd->var) > 0 && getVar(nd->var) <= hcs->varnum);
    return nd;
}

hc_node_t *follow_path(hc_support_t *hcs, hc_node_t *n, flexarray_t *c) {
    hc_node_t *nd = n, *new_root = NULL, *nlast = NULL;
    uint *ca = getFlexArray(c);
    uint elim_sz, pbranch_sz;
    elim_sz = stackSize(hcs->elim);
    pbranch_sz = stackSize(hcs->pbranch);

    while (TRIM(nd)) {
#ifdef DEBUG
        uint dbg_v = getVar(nd->var);
        assert(dbg_v <= hcs->varnum && dbg_v > 0);
#endif
        if (!hc_is_bnode(nd)) {
            stack_t *update = NULL;
            int i = find_in_certificate(hcs, nd->var);
            if (i < 0) { // node not found
                update = hcs->elim;
            } else if (ca[i] != nd->var) { // potential new branching node
                assert(i >= 0);
                markStamp(hcs, (uint)i);
                update = hcs->pbranch;
            }
            if (update) {
                stackPush(update, (Item)nd);
                nd = nd->nxt;
            } else { // literal found in certificate
                hc_node_t *xn = nd->nxt;
                assert(i >= 0);
                markStamp(hcs, (uint)i);
                if (nlast) {
                    append(nlast, nd);
                    nlast = nd;
                } else if (!new_root) {
                    nlast = new_root = nd;
                    new_root->nxt = TERM_ND;
                    assert(!new_root->neg);
                }
                nd = xn;
            }
        } else {
            hc_node_t *bnd = follow_branch(hcs, nd, c);
            // follow_branch can return either the same node or NULL
            assert(bnd == nd || !bnd);
            if (nlast) {
                hc_link(nlast, bnd);
            } else if (!new_root && bnd) {
                nlast = new_root = bnd;
            }
            goto cleanup;
        }
    }
    
    if (!stackEmpty(hcs->pbranch) && !hcs->too_branchy) {
        uint pbnodes, i, sz = getFlexArraySize(c);
        hc_node_t *first[2] = {NULL, NULL}, *last[2] = {NULL, NULL};
        // nd is a new branching node
        nd = pickBestBnode(hcs);
        // pbnodes is a number of variables that have a match in
        // the certificate, but literal is not matching (potentially
        // branching nodes)
        pbnodes = enlistStacks(hcs, &first[0], &last[0]);
        invListCopy(hcs, first[0], pbnodes, &first[1], &last[1]);
        
        // handle literals present in certificate and not in
        // current hcube
        for (i = 0; i < sz; i++) {
            if (!getStamp(hcs, i)) {
                hc_node_t *xn = getHcNode(hcs, ca[i]);
                if (last[1]) {
                    append(last[1], xn);
                    last[1] = xn;
                } else {
                    first[1] = last[1] = xn;
                }
            }
        }
        // construct new branching node
        if (first[0] || first[1]) {
            hc_node_t **pos = LSB(nd->var) ? &(nd->nxt) : &(nd->neg);
            hc_node_t **neg = LSB(nd->var) ? &(nd->neg) : &(nd->nxt);
            
            if (first[0]) {
                *pos = first[0];
            } else {
                *pos = TERM_ND;
            }
            if (first[1]) {
                *neg = first[1];
            } else {
                *neg = TERM_ND;
            }
            
            // update number of nodes
            hcs->addedNodes--;
            hcs->addedBnodes++;
            // link new branching node
            // and update root if necessary
            if (nlast) {
                hc_link(nlast, nd);
            } else if (!new_root) {
                new_root = nd;
            }
        } else {
            // return unused branching node to the stack
            stackPush(hcs->pbranch, (Item)nd);
        }
    }

cleanup:
    while (stackSize(hcs->elim) > elim_sz) {
        hc_node_t *x = (hc_node_t *)stackPop(hcs->elim);
        freeOneHcNode(hcs, x);
    }
    while (stackSize(hcs->pbranch) > pbranch_sz) {
        hc_node_t *x = (hc_node_t *)stackPop(hcs->pbranch);
        freeOneHcNode(hcs, x);
    }
    return TERMINATE(new_root);
}

void init_ctable(hc_support_t *hcs, flexarray_t *c) {
    uint *ca = getFlexArray(c);
    uint i, csz = getFlexArraySize(c);
    incrTimeStamp(hcs->ctable);
    for (i = 0; i < csz; i++) {
        uint lit = ca[i];
        uint var = getVar(lit);
#ifdef DEBUG
        uint res = 
#endif
        writeTable(hcs->ctable, var, LSB(lit));
        hcs->indices[var] = i;
#ifdef DEBUG
        assert(res == WRITE_SUCCESS);
#endif
    }
}

// hc - current hypercube, c - new certificate
void common_path(hc_support_t *hcs, hc_anchor_t *hc, flexarray_t *c) {
    hc_node_t *n = hc->nd;
    assert(stackEmpty(hcs->elim));
    assert(stackEmpty(hcs->pbranch));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);

    if ((uint)n < 2) { // if n is a termination node or NULL
        if (hc->has_hcube) {
            assert(hcs->addedNodes == 0);
            assert(hcs->addedBnodes == 0);
            return;
        } else {
            // just initalize it to certificate
            hc_node_t *x, *last = NULL, *first = NULL;
            uint *ca = getFlexArray(c);
            uint i, csz = getFlexArraySize(c);
            for (i = 0; i < csz; i++) {
                x = getHcNode(hcs, ca[i]);        
                if (last) {
                    append(last, x);
                    last = x;
                } else {
                    first = last = x;
                }
            }
            hc->nd = first;
            hc->has_hcube = TRUE;
        }
    } else { // add a new certificate to BCT
#ifdef DEBUG
        uint i;
#endif
        clearStamps(hcs, getFlexArraySize(c));
#ifdef DEBUG
        for (i = 0; i < getFlexArraySize(c); i++) {
            assert(hcs->stamps[i] == 0);
        }
#endif
        init_ctable(hcs, c);
        hcs->too_branchy = hc_too_branchy(hcs, hc);
        if (hc_is_bnode(n)) {
            hc->nd = TRIM(follow_branch(hcs, hc->nd, c));
        } else {
            hc->nd = TRIM(follow_path(hcs, hc->nd, c));
        }
        resetHcStacks(hcs);
    }
    hc_update_cnts(hcs, hc);
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
}

// computes a deep copy of BCT *n according to what has been
// seen on the path in the other BCT, to which *n is appended
hc_node_t *hc_append(hc_support_t *hcs, hc_node_t *n) {
    hc_node_t *x = n;
    hc_node_t *first = TERM_ND, *last = NULL;
    if ((uint)n < 2) {
        return n;
    } else {
        do {
            uint lit = x->var;
            uint var = getVar(x->var);
            int i = checkTable(hcs->ctable, var);
            if (hc_is_bnode(x)) { // branch node
                hc_node_t *xn;
                if (i < 0) { // no such var in certificate
                    hc_node_t *nxt = hc_append(hcs, x->nxt);
                    hc_node_t *neg = hc_append(hcs, x->neg);
                    if (!nxt && !neg) { // conflict
                        freeHcNode(hcs, first);
                        return NULL;
                    } else if (nxt == neg) { // redundant node
                        assert(nxt == TERM_ND);
                        return first;
                    }
                    
                    xn = getHcNode(hcs, lit);

                    if (last) {
                        assert(!hc_is_bnode(last));
                        append(last, xn);
                    } else {
                        first = last = xn;
                    }

                    if (!nxt) {
                        assert((uint)neg >= 1);
                        xn->var &= ~1;
                        xn->nxt = neg;
                    } else if (!neg) {
                        assert((uint)nxt >= 1);
                        xn->nxt = nxt;
                        xn->var |= 1;
                    } else { // create a branching node
                        xn->nxt = nxt;
                        xn->neg = neg;
                    }
                    return first;
                } 
                
                if (i) { // TRUE lit existing on the path
                    x = x->nxt;
                } else { // FALSE lit existing on the path
                    x = x->neg;
                }
            } else { // literal node
                if (i < 0) { // no such literal on the current path
                    hc_node_t *xn = getHcNode(hcs, lit);
                    if (last) {
                        append(last, xn);
                        last = xn;
                    } else {
                        first = last = xn;
                    }    
                    x = x->nxt;
                } else if ((uint)i == LSB(lit)) { // double literal
                    x = x->nxt;
                } else { // conflict
                    freeHcNode(hcs, first);
                    return NULL;
                }
            }
            assert((uint)x >= 1);
        } while (x != TERM_ND);
        return first;
    }
}

// computes the join (intersection) of node *n and hcs->thc
hc_node_t *hc_do_join(hc_support_t *hcs, hc_node_t *n) {
    hc_node_t *x = n;
    stack_t *seen = hcs->elim;
    if ((uint)n < 2) {
        // leaf node reached - append BCT stored in hcs->thc
        return hc_append(hcs, hcs->thc);
    } else {
        do {
			uint lit, var;
            assert((uint)x > 1);
            lit = x->var;
            var = getVar(lit);
            assert(var <= hcs->varnum && var > 0);
            if (hc_is_bnode(x)) { // branch node
                uint ptr = stackSize(seen);
                
                // follow positive branch
                stackPush(seen, (Item)var);
                (void)writeTable(hcs->ctable, var, (uint)1);
                x->nxt = hc_do_join(hcs, x->nxt);
                while (stackSize(seen) > ptr) {
                    uint v = (uint)stackPop(seen);
                    resetTableElem(hcs->ctable, v);
                }

                // follow negative branch
                stackPush(seen, (Item)var);
                (void)writeTable(hcs->ctable, var, (uint)0);
                x->neg = hc_do_join(hcs, x->neg);
                while (stackSize(seen) > ptr) {
                    uint v = (uint)stackPop(seen);
                    resetTableElem(hcs->ctable, v);
                }

                if (!x->nxt && !x->neg) { // conflict
                    freeHcNode(hcs, n);
                    return NULL;
                } else if (!x->nxt) {
                    x->nxt = x->neg;
                    x->neg = NULL;
                    x->var &= ~1;
                } else if (!x->neg) {
                    x->var |= 1;
                }
                assert((uint)x->nxt >= 1);
                return n;
            } else { // literal node
                stackPush(seen, (Item)var);
                (void)writeTable(hcs->ctable, var, LSB(lit));
                assert((uint)x->nxt >= 1);
                if (x->nxt == TERM_ND) {
                    hc_node_t *xn = hc_append(hcs, hcs->thc);
                    if (!xn) {
                        freeHcNode(hcs, n);
                        return NULL;
                    } else {
                        assert((uint)xn >= 1);
                        x->nxt = xn;
                    }
                    return n;
                } else {
                    x = x->nxt;
                }
            }    
        } while (TRUE);
    }
}

// puts all pairs [parent branch] on the stack. those pairs
// are used for simplification (elimination of equal literals
// in adjacent branches)
void hc_find_branches(hc_support_t *hcs, hc_node_t *n) {
    hc_node_t *x = n, *branch_pre = NULL;
    stack_t *todo = hcs->pbranch;
    stack_t *branches = hcs->elim;
    uint sgn = FALSE;

    assert(stackEmpty(todo));
    assert(stackEmpty(branches));

    do {
        while ((uint)x > 1) {
#ifdef DEBUG
            uint dbg_v = getVar(x->var);
            assert(dbg_v <= hcs->varnum && dbg_v > 0);
#endif
            if (hc_is_bnode(x)) {
                stackPush(todo, (Item)x);
                // push the previous node to stack. sgn is necessary
                // to determine which branch leads to the pushed x
                // branch node.
                stackPush(branches, (Item)((uint)branch_pre | sgn));
                // push branch node itself
                stackPush(branches, (Item)x);
                branch_pre = x;
                sgn = FALSE;
                x = x->neg;
            } else {
                branch_pre = x;
                sgn = TRUE;
                x = x->nxt;
            }
        }
        if (!stackEmpty(todo)) {
            x = (hc_node_t *)stackPop(todo);
            assert(hc_is_bnode(x));
            branch_pre = x;
            sgn = TRUE;
            x = x->nxt;
        } else {
            return;
        }
    } while (TRUE);
}

hc_node_t *hc_do_simplify(hc_support_t *hcs, hc_node_t *n) {
    stack_t *branches = hcs->elim;
    hc_node_t *first = n;
    assert(stackEmpty(branches));
    // push all branches to the stack in depth first manner
    hc_find_branches(hcs, n);

    while (!stackEmpty(branches)) {
        hc_node_t *bnd, *bnd_pre, *x_pre, *x;
        uint bnd_pre_branch;
        incrTimeStamp(hcs->ctable);
        
        x_pre = bnd = (hc_node_t *)stackPop(branches);

        assert((uint)bnd > 1);
        assert(hc_is_bnode(bnd));
        assert(bnd->nxt != bnd->neg);

        bnd_pre_branch = (uint)stackPop(branches);
        bnd_pre = TRIM(bnd_pre_branch);
        bnd_pre_branch &= 1;

        assert(!bnd_pre || bnd_pre->nxt == bnd || bnd_pre->neg == bnd);
        assert(bnd->nxt != bnd->neg);

#ifdef DEBUG
        if ((uint)bnd_pre > 1) {
            uint tmp_var = getVar(bnd_pre->var);
            assert(tmp_var > 0 && tmp_var <= hcs->varnum);
            assert((!bnd_pre_branch) || bnd_pre->nxt == bnd);
            assert((bnd_pre_branch) || bnd_pre->neg == bnd);
        }
#endif

        // enter the FALSE branch into the unique table
        x = bnd->neg;
        while (x != TERM_ND && !hc_is_bnode(x)) {
            uint lit = x->var;
            uint var = getVar(lit);
            assert(var > 0 && var <= hcs->varnum);
            // write sign into the table
            writeTable(hcs->ctable, var, LSB(lit));
            // enter previous node in indices table
            hcs->indices[var] = (uint)x_pre;
            x_pre = x;
            x = x->nxt;
            assert(x != NULL);
        }
        assert(bnd->nxt != bnd->neg);

        // scan the TRUE branch for literals also
        // present in the false branch until the next
        // branching node or termination node (encoded as 1)
        // is reached
        x = bnd->nxt;
        x_pre = bnd;
        while ((uint)x > 1 && !hc_is_bnode(x)) {
            uint lit = x->var;
            uint var = getVar(lit);
			int i;

            assert(x_pre->nxt == x);
            assert(var > 0 && var <= hcs->varnum);
            assert(bnd->nxt != bnd->neg);

            i = checkTable(hcs->ctable, var);
            if ((uint)i != LSB(lit)) { // no match in the adjacent branch
                x_pre = x;
                x = x->nxt;
            } else { // equal literal found in adjacent branches
#ifdef DEBUG
                uint tmp_var;
#endif
                hc_node_t *neg_del, *neg_del_pre, *neg_del_nxt;
                hc_node_t *pos_pre = x_pre, *pos_nxt = x->nxt;
                // get the parent of the corresponding equal
                // literal node in the adjacent (negative) branch
                neg_del_pre = (hc_node_t *)hcs->indices[var];
                // note: the node to be deleted must have a
                // parent, otherwise it couldn't be a negative
                // branch of any node

                assert((uint)neg_del_pre > 1);
                assert(pos_pre != pos_nxt);
#ifdef DEBUG
                tmp_var = getVar(neg_del_pre->var);
                assert(tmp_var > 0 && tmp_var <= hcs->varnum);
#endif
                // the node to be deleted is determined according to
                // the type of its parent
                neg_del = (hc_is_bnode(neg_del_pre)) ? neg_del_pre->neg :
                    neg_del_pre->nxt;

                assert((uint)neg_del > 1);
                assert(!hc_is_bnode(neg_del));
                assert(neg_del->var == lit);
                assert(bnd->nxt != bnd->neg);
                assert(pos_pre != pos_nxt);

                // if the node to be deleted has a child,
                // update its parent in the table
                neg_del_nxt = neg_del->nxt;
                if ((uint)neg_del_nxt > 1) {
                    uint nxt_neg_var = getVar(neg_del_nxt->var);
                    assert(nxt_neg_var > 0 && nxt_neg_var <= hcs->varnum);
                    hcs->indices[nxt_neg_var] = (uint)neg_del_pre;
                }
                freeOneHcNode(hcs, neg_del);
                neg_del = NULL;

                assert(bnd->nxt != bnd->neg);
                assert(pos_pre != pos_nxt);

                // short-connect neg_del node correspoding to x in 
                // negative branch
                if (hc_is_bnode(neg_del_pre)) {
                    assert(neg_del_pre == bnd);
                    assert((uint)neg_del_nxt >= 1);
                    neg_del_pre->neg = neg_del_nxt;
                } else {
                    hc_link(neg_del_pre, neg_del_nxt);
                }

                assert(bnd->nxt != bnd->neg);
                assert(pos_pre != pos_nxt);

                // cut out x node from positive branch
                hc_link(pos_pre, pos_nxt);
                assert(bnd->nxt != bnd->neg || bnd->nxt == TERM_ND);
                assert(pos_pre != pos_nxt);
                assert(pos_pre != pos_nxt);

                // insert x node in the right place
                x->nxt = bnd;
                if (!bnd_pre) {
                    assert(first != NULL);
                    hc_link(x, first);
                    first = x;
                } else if (bnd_pre_branch) {
                    bnd_pre->nxt = x;
                } else {
                    assert(!bnd_pre_branch);
                    bnd_pre->neg = x;
                } 

                // remove leaf branching node
                if (bnd->neg == bnd->nxt) {
                    assert((uint)bnd->neg == 1);
                    freeOneHcNode(hcs, bnd);
                    x->nxt = TERM_ND;
                }
                x = pos_nxt;
            }
        }
    }
    assert(stackEmpty(hcs->elim));
    assert(stackEmpty(hcs->pbranch));
    return first;
}

void hc_simplify(hc_support_t *hcs, hc_anchor_t *ha) {
    assert(stackEmpty(hcs->elim));
    ha->nd = TRIM(hc_do_simplify(hcs, ha->nd));
    hc_update_cnts(hcs, ha);
    assert(stackEmpty(hcs->elim));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
}

uint hc_join(hc_support_t *hcs, hc_anchor_t **hc) {
    assert(stackEmpty(hcs->elim));
    if ((uint)hc[0]->nd < 2) {
        SWAP((uint *)hc[0], (uint *)hc[1]);
        return FALSE;
    } else if ((uint)hc[1]->nd < 2) {
        return FALSE;
    } else if (hc[0]->num_bnodes > hc[1]->num_bnodes) {
        SWAP((uint *)hc[0], (uint *)hc[1]);
    }
    incrTimeStamp(hcs->ctable);
    hcs->thc = hc[1]->nd;
    hc[0]->nd = TRIM(hc_do_join(hcs, hc[0]->nd));
    assert(!hc[0]->nd || (uint)hc[0]->nd > 1);
    stackReset(hcs->elim);
    assert(stackEmpty(hcs->elim));
    assert(stackEmpty(hcs->pbranch));
    hcs->thc = NULL;
    if (!hc[0]->nd) {
        hcs->conflict = TRUE;
    }
    resetHcube(hcs, hc[1]);
    return TRUE;
}

// ihc - inherited hypercube, nhc - new hypercube
uint path_intersect(hc_support_t *hcs, hc_anchor_t **hc) {
    uint needs_simplification = hc_join(hcs, hc);
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
    if (!hcs->conflict) {
        if (needs_simplification) {
            hc_simplify(hcs, hc[0]);
        }
        assert(hcs->addedNodes == 0);
        assert(hcs->addedBnodes == 0);
        return FALSE; // no conflict
    } else {
        hcs->conflict = FALSE;
        assert(hcs->addedNodes == 0);
        assert(hcs->addedBnodes == 0);
        return TRUE; // conflict found
    }
}

// returns TRUE if there would be any change after x is
// appended to the current branch, meaning that x is not
// subsumed, FALSE otherwise.
uint hc_try_append(hc_support_t *hcs, hc_node_t *x) {
    stack_t * todo = hcs->pbranch;
    assert(stackEmpty(todo));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
    assert(x != NULL);
    assert(x != TERM_ND);
    do {
        if (x == TERM_ND) {
            if (!stackEmpty(todo)) {
                x = (hc_node_t *)stackPop(todo);
            } else {
                break;
            }
        } else {
            uint lx = x->var;
            uint vx = getVar(lx);
            int ex = checkTable(hcs->ctable, vx);
            if (!hc_is_bnode(x)) { // literal node
                if (ex < 0 || (uint)ex != LSB(lx)) {
                    stackReset(todo);
                    return TRUE;
                }
                x = x->nxt;
            } else { // branch node
                hc_node_t *nxt = x->neg;
                assert(hc_is_bnode(x));
                if (ex < 0) {
                    stackPush(todo, (uint)x->nxt);
                } else if (ex == 1) {
                    nxt = x->nxt;
                }
                x = nxt;
            }
        }
    } while (TRUE);
    
    assert(stackEmpty(todo));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
    return FALSE;
}

// returns TRUE if n1 is subsumed by n2, FALSE otherwise
uint check_subsumption(hc_support_t *hcs, hc_node_t *n1, hc_node_t *n2) {
    stack_t *todo = hcs->elim;
    hc_node_t *x = n1;
    assert(stackEmpty(todo));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);

    do { 
        assert(x != NULL);
        assert(n2 && !(n2 == TERM_ND));
        if (x == TERM_ND) {
            // hc_try_append returns true if there would be any change
            // (either some nodes get appended, or a branch gets deleted)
            if (hc_try_append(hcs, n2)) {
                stackReset(todo);
                assert(hcs->addedNodes == 0);
                assert(hcs->addedBnodes == 0);
                return FALSE;
            }
        } else if (!hc_is_bnode(x)) {
            uint lx = x->var;
            uint vx = getVar(lx);
            assert(vx > 0 && vx <= hcs->varnum);
            stackPush(todo, (Item)x);
            (void)writeTable(hcs->ctable, vx, LSB(lx));
            x = x->nxt;
            continue;
        } else {
            uint vx = getVar(x->var);
            assert(hc_is_bnode(x));
            assert(vx > 0 && vx <= hcs->varnum);
            stackPush(todo, (Item)((uint)x | 1));
            (void)writeTable(hcs->ctable, vx, (uint)0);
            x = x->neg;
            continue;
        }
        if (!stackEmpty(todo)) {
            uint nn = (uint)stackPop(todo);
            hc_node_t *nx = (hc_node_t *)TRIM(nn);
            uint vx = getVar(nx->var);
            assert(vx > 0 && vx <= hcs->varnum);
            resetTableElem(hcs->ctable, vx);
            if (nn & 1) {
                x = nx->nxt;
                stackPush(todo, (uint)nx);
                (void)writeTable(hcs->ctable, vx, (uint)1);
            }
        } else {       
            break;
        }
    } while (TRUE);
    
    assert(stackEmpty(todo));
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
    return TRUE;
}

// returns:
// TRUE if n1 is subsumed by n2, so n2 needs to be explored first
// FALSE if n2 is subsumed by n1, so n1 needs to be explored first
// NONE if there's no subsumption relation.
// the function should be called as 
// subsumption(hcs, bnd->neg, bnd->pos)
uint subsumption(hc_support_t *hcs, hc_node_t *n1, hc_node_t *n2) {
    uint is_bnd1, is_bnd2;
    assert(!(n1 == TERM_ND && n2 == TERM_ND));
    assert(n1 && n2);
    assert(n1 != n2);
    assert(hcs->addedNodes == 0);
    assert(hcs->addedBnodes == 0);
    if (n1 == TERM_ND) {
        return FALSE;
    } else if (n2 == TERM_ND) {
        return TRUE;
    } 
    is_bnd1 = hc_is_bnode(n1);
    is_bnd2 = hc_is_bnode(n2);
    incrTimeStamp(hcs->ctable);
    if (!is_bnd1 && !is_bnd2) {
        // there's no subsumption relation if the first nodes
        // of both BCTs are literal nodes
        return NONE;
    } else if (is_bnd1 && is_bnd2) {
        // try to establish a subsumption relation both ways
        return (check_subsumption(hcs, n1, n2)) ? TRUE :
            (check_subsumption(hcs, n2, n1)) ? FALSE : NONE;
    } else if (is_bnd1) {
        // check whether n2 is subsumed by n1
        return (check_subsumption(hcs, n2, n1)) ? FALSE : NONE;
    } else {
        assert(is_bnd2);
        // check whether n1 is subsumed by n2
        return (check_subsumption(hcs, n1, n2)) ? TRUE : NONE;
    }
}

void resetHcube(hc_support_t *hcs, hc_anchor_t *ha) {
    freeHcNode(hcs, ha->nd);
    ha->nd = NULL;
    ha->has_hcube = FALSE;
    resetHcSupport(hcs);
}

/*
 * Revision history:
 *
 * $Log: hcube.h,v $
 * Revision 1.10  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.9  2005/01/30 13:21:33  root
 * Fixed minor performance bug in storing flags.
 *
 * Revision 1.8  2005/01/01 10:07:01  root
 * Fixed a bug in storing time stamps.
 *
 * Revision 1.7  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.5  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.4  2004/12/11 04:46:12  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.3  2004/12/09 18:51:33  root
 * Minor modification.
 *
 * Revision 1.2  2004/12/09 18:50:26  root
 * Added CVS headers.
 *
 *
 */
