/* Author: Domagoj Babic 2004.
 *
 * $Id: hsat.c,v 1.7 2005/02/21 22:32:30 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#include "hsat.h"
#include "simplify.h"

void usage(void) {
    printf("\n Format: hsat input.cnf [options]\n\n");
    printf(" Options:\n");
    printf(" -h     \tprints this short help\n");
    printf(" -f file\twrite result to file\n");
    printf(" -c     \tverify the result\n");
    printf(" -m [opt]\tmodel printout format [VECT | RAW]\n");
    printf("        \tVECT - vector format \"v -1 2 3 -4...\", default\n");
    printf("        \tRAW - report only SAT | UNSAT\n");
    printf(" -t [sec]\ttimeout in seconds\n");
    printf(" -noeqv \tskip preprocessing of equivalences\n");
    printf(" -clust \tdetect clusters in the problem\n");
    printf(" -cperc \tmin cluster size [1..100], default 5\n");
    printf("\n");
    printf(" Example:\n");
    printf(" hsat bm1.cnf -f bm1.res -c -m VECT\n");
    printf(" would take bm1.cnf as input, write the result to \
bm1.res,\n");
    printf(" and verify the model (if satisfiable). The solver would\n");
    printf(" report result in vector format\n");
    printf("\n");
    printf(" WARRNING: clustering requires MCL package!\n");
    exit(EXIT_UNKNOWN);
}

int main (int argc, char *argv[]) {
    clock_t start = clock(), end = 0;
    flexarray_t *cnf = NULL;
    preproc_t *p = NULL;
    stats_t *stats = xcalloc(1, sizeof(stats_t));
	int i, exit_status;
    uint cluster_perc = 5, vnum = 0, cnum = 0;
    satm_t *sm = NULL;
    char toVerify = FALSE, verbose = FALSE, rstats = TRUE;
    char mformat = VECT, *ofile = NULL;
    char eqv = TRUE, cluster = FALSE;

    if (argc < 2) {
        error("Usage: sat CNF_file [options (see with -h)]");
    }
    
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-h")) {
            usage();
        } else if (!strcmp(argv[i], "-t")) {
            uint tout;
            i++;
            if (i >= argc) {
                error("-t option should be followed by a \
number of seconds!");
            }
            tout = (uint)atol(argv[i]);
            if (tout == 0) {
                error("Invalid timeout value!");
            }
            end = tout*CLOCKS_PER_SEC + start;
        } else if (!strcmp(argv[i], "-f")) {
            i++;
            if (i >= argc) {
                error("-f option should be followed by the \
file name!");
            }
            ofile = argv[i];
        } else if (!strcmp(argv[i], "-noeqv")) {
            eqv = FALSE;
        } else if (!strcmp(argv[i], "-clust")) {
            cluster = TRUE;
        } else if (!strcmp(argv[i], "-cperc")) {
            uint cp;
            i++;
            cp = (uint)atol(argv[i]);
            if (cluster_perc < 1 || cluster_perc > 100) {
                warrn("Invalid min cluster percentage [1..100]");
            } else {
                cluster_perc = cp;
            }
        } else if (!strcmp(argv[i], "-v")) {
            verbose = TRUE;
            note("Verbosity levels not yet implemented!");
        } else if (!strcmp(argv[i], "-c")) {
            toVerify = TRUE;
        } else if (!strcmp(argv[i], "-ns")) {
            rstats = FALSE;
        } else if (!strcmp(argv[i], "-m")) {
            i++;
            if (i >= argc) {
                error("-m option should be followed by a \
model output type [VECT | RAW]\n");
            }
            if (!strcmp(argv[i], "VECT")) {
                mformat = VECT;
            } else if (!strcmp(argv[i], "RAW")) {
                mformat = RAW;
            } else {
                warrn("Unsupported model format!");
            }
        }
    }

    pagelock();
    cnf = parse(argv[1], &vnum, &cnum);
    stats->clOrigClauses = cnum;
    stats->varNum = vnum;
    
    p = initPreproc(vnum, cnf, stats, eqv, cluster, cluster_perc);
    simplify(p);

    // if not simplified to FALSE
    if (p->sat_status == NONE) {
        flexarray_t *aq; //, *coe = getCOE(p);
        sm = newManager(vnum, cnum, p->elimvars, p->depvars, 
                p->clusters, stats);
        initManager(sm, cnf, p->ecnf, p->pri, end);
        //freeFlexArray(coe);
        partiallyFreePreproc(p);
        printf("c Preprocessing done in %f [s]\n", 
                ((double)(clock()-start))/CLOCKS_PER_SEC);
        search(sm);
        p->sat_status = sm->status;
        aq = getAQ(sm->aq);
        postprocess(p, aq, mformat, ofile, argv[1], start);
        freeFlexArray(aq);
        if (sm->status == TRUE && toVerify) {
            verify(p, argv[1]);
        }
        if (sm->status == TRUE) {
            exit_status = EXIT_SATISFIABLE;
        } else if (sm->status == NONE) {
            exit_status = EXIT_UNKNOWN;
        } else {
            exit_status = EXIT_UNSATISFIABLE;
        }
        freeManager(sm);
    } else {
        exit_status = EXIT_UNSATISFIABLE;
        postprocess(p, NULL, mformat, ofile, argv[1], start);
        partiallyFreePreproc(p);
    }
#ifdef COLLECT_STATS
    if (rstats) {
        reportStats(stats, ofile);
    }
#endif
    free(stats);
    freePreproc(p);
    exit(exit_status);
}

/*
 * Revision history:
 *
 * $Log: hsat.c,v $
 * Revision 1.7  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.6  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.4  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.3  2004/12/09 18:52:09  root
 * Minor modification.
 *
 * Revision 1.2  2004/10/24 00:05:40  root
 * Comment added.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
